package cp.pairing;

import cp.*;
import java.security.acl.*;
import java.util.Collections;
import java.util.Vector;


public class NordicSchweizer extends PairingSystem {

  public NordicSchweizer(Tournament tournament) {
    super(tournament);
    initialize();
  }

  protected Games pair(Players players) throws ChessPairerException {
    // Divide players in score groups
    Vector<Players> groups = createGroups(players);
    return pair(groups, false);
  }

  private Games pair(Vector<Players> groups, boolean firstGameRemoved) throws ChessPairerException {
    Games newGames = new Games(tournament);
    // Pair each group
    for (int i=0; i<groups.size(); i++) {
      Players group = groups.get(i);
      if (i == 0 && group.size() == 1 && !firstGameRemoved) {
        // Pair according to 14B:3d
        Game game = pair_3d(group.get(0), groups, newGames);
        if (game != null) {
          remove(game.getWhitePlayer(), groups);
          remove(game.getBlackPlayer(), groups);
          Games games = pair(groups, true);
          if (games == null) {
            return null;  
          }
          newGames.addAll(games);
          return newGames;
        }
        else {
          return null;
        }
      }
      else if (group.size() %2 == 1) {
        // Pair according to 14B:3e
        if (!movePlayer(group, groups, i)) {
          // That didn't work. Concatenate the last two groups!
          if (!concatenateLastTwoGroups(groups)) {
            return null;
          }
        }
        return pair(groups, firstGameRemoved);
      }
      else if (group.size() %2 == 0) {
        // Pair according to 14B:3f
        Games games = pair_3f(group);
        if (games.isEmpty()) {
          // Pairing failed. Apply 14B:2i
          if (!movePlayer(group, groups, i)) {  // Note: Even if only one player is moved, it will be called again because the group will be odd!
            // That didn't work. Concatenate the last two groups!
            if (!concatenateLastTwoGroups(groups)) {
              return null;
            }
          }
          return pair(groups, firstGameRemoved);
        }
        newGames.addAll(games);
      }
    }
    return newGames;
  }

  /**
   * Pairs a group according to 14B:3f. Players must be an even number.
   */
  private Games pair_3f(Players players) throws ChessPairerException {
    // Divide players in two groups 
    // Loop until pairing is ok
    Games games;
    int swapPlayers = 0;
    do { 
      Players upper = new Players(tournament.getSettings());
      Players lower = new Players(tournament.getSettings());
      upper.addAll(players.subList(0, players.size()/2));
      lower.addAll(players.subList(players.size()/2, players.size()));
      for (int i=0; i<swapPlayers; i++) {
        Player playerU = upper.get(upper.size()-i-1);
        Player playerL = lower.get(i);
        upper.add(upper.size()-(swapPlayers-i), playerL);
        lower.add(swapPlayers-i-1, playerU);
        upper.remove(playerU);
        lower.remove(playerL);
      }
      games = pair_3f(upper, lower);
      if (games.size() == 0) {
        swapPlayers++;        
      }
    } while (games.size() == 0 && swapPlayers < players.size()/2);
    return games;
  }
  
  private Games pair_3f(Players upper, Players lower) throws ChessPairerException {
    Games games = new Games(tournament);
    Games forbiddenGames = new Games(tournament);
    // For each player:
    for (int i=0; i < upper.size(); i++) {
      Player a = upper.get(i);
      // for each remaining player in lower:
      for (int j=0; j < lower.size(); j++) {
        Player b = lower.get(j);
        // if the opponent does not meet anyone and they has not met each other
        if (!a.hasMet(b) && !isBarred(a,b)) {
          // if pair is not in forbidden list
          if (!hasMet(a, b, forbiddenGames)) {
            Game game;
            // create game and lot the rest
            if (b.getWhiteCount() > a.getWhiteCount()) {
              game = new Game(games, a, b);
            }
            else if (b.getWhiteCount() < a.getWhiteCount()) {
              game = new Game(games, b, a);
            }
            else if (tournament.getRounds().size() % 2 == 1) {  // This round is not created yet. Odd means this will be an even roundnr.
              game = new Game(games, a, b);
            }
            else {
              game = new Game(games, b, a);
            }
            int aind = upper.indexOf(a);
            int bind = lower.indexOf(b);
            upper.remove(a);
            lower.remove(b);
            Games rest = pair_3f(upper, lower);
            upper.add(aind, a);
            lower.add(bind, b);

            if (rest.size() == 0 && upper.size() > 1 && lower.size() > 1) {
              // Can not pair. Move game to forbidden games
              forbiddenGames.add(game);
            }
            else {
              games.add(game);
              games.addAll(rest);
              return games;
            }
          }
        }
      }
    }
    return games;
  }

  private Vector<Players> createGroups(Players players) {
    Vector<Players> ans = new Vector<Players>();
    // First sort players according to score and lot nr.
    Round round = null;
    if (tournament.getRounds().size() > 0) {
      round = tournament.getRounds().lastElement();
    }
    Collections.sort(players, new ScoreLotnrComparator(round));
    double lastScore = Double.NaN;
    Players group = new Players(tournament.getSettings());
    for (int i=0; i<players.size(); i++) {
      Player player = players.get(i);
      double score = player.getPairingScore(round);
      if (score != lastScore) {
        lastScore = score;
        if (group.size() > 0) {
          ans.add(group);
        }
        group = new Players(tournament.getSettings());
        group.add(player);
      }
      else {
        group.add(player);
      }
    }
    ans.add(group);
    return ans;
  }

  /**
   * Pairs a group according to 14B:3d. 
   */
  private Game pair_3d(Player a, Vector<Players> groups, Games games) throws ChessPairerException {
    for (int i=0; i<groups.size(); i++) {
      Players group = groups.get(i);
      for (int j=0; j<group.size(); j++) {
        Player b = group.get(j);
        if (b != a && !games.hasPlayer(b) && !a.hasMet(b) && !isBarred(a,b)) {
          Game game;
          // create game and lot the rest
          if (b.getWhiteCount() > a.getWhiteCount()) {
            game = new Game(games, a, b);
          }
          else if (b.getWhiteCount() < a.getWhiteCount()) {
            game = new Game(games, b, a);
          }
          else if (tournament.getRounds().size() % 2 == 1) {  // This round is not created yet. Odd means this will be an even roundnr.
            game = new Game(games, a, b);
          }
          else {
            game = new Game(games, b, a);
          }
          games.add(game);
          return game;
        }
      }
    }
    return null;
  }

  private boolean concatenateLastTwoGroups(Vector<Players> groups) {
    if (groups.size() > 1) {
      Players last = groups.lastElement();
      Players nextLast = groups.get(groups.size()-2);
      nextLast.addAll(last);
      groups.remove(last);
      return true;
    }
    return false;
  }
  
  /**
   * Moves the best player from a lower group to a specified group. If the group is the last group, the last two groups are conatenated.
   * If concatenation fails, it returns false (and pairing cannot be done).
   */
  private boolean movePlayer(Players currentPlayers, Vector<Players> groups, int i) {
    if (i+1 >= groups.size()) {
      // This is the last group. Concatenate it with the previous instead.
      return concatenateLastTwoGroups(groups);
    }
    else {
      Players group = groups.get(i);
      // Find the next player that has not met all players in group 'group'
      for (int j=i+1; j<groups.size(); j++) {
        Players nextGroup = groups.get(j);
        for (int k=0; k<nextGroup.size(); k++) {
          Player player = nextGroup.get(k);
          if (!hasMetAll(player, currentPlayers)) {
            group.add(player);
            nextGroup.remove(player);
            if (nextGroup.isEmpty()) {
              groups.remove(nextGroup);
            }
            return true;
          }
        }
      }
    }
    return false;
  }

  private void remove(Player player, Vector groups) {
    for (int i=0; i<groups.size(); i++) {
      Vector group = (Vector) groups.get(i);
      if (group.remove(player)) {
        if (group.isEmpty()) {
          groups.remove(group);
          i--;
        }
      }
    }
  }

  private boolean hasMetAll(Player player, Players players) {
    for (int i=0; i<players.size(); i++) {
      Player opponent = players.get(i);
      if (!player.hasMet(opponent)) {
        return false;
      }
    }
    return true;
  }
}
